import os
from typing import Any, Dict

from valohai_yaml.objs.config import Config
from valohai_yaml.objs.input import Input, KeepDirectories
from valohai_yaml.objs.parameter import Parameter
from valohai_yaml.objs.step import Step

from valohai.consts import DEFAULT_DOCKER_IMAGE
from valohai.internals.parsing import parse

ParameterDict = Dict[str, Any]
InputDict = Dict[str, str]


def generate_step(
    *,
    relative_source_path: str,
    step: str,
    image: str,
    parameters: ParameterDict,
    inputs: InputDict,
) -> Step:
    config_step = Step(
        name=step,
        image=image,
        command=[
            "pip install -r requirements.txt",
            "python %s {parameters}" % relative_source_path,
        ],
    )

    for key, value in parameters.items():
        config_step.parameters[key] = Parameter(
            name=key,
            type=get_parameter_type_name(key, value),
            default=value,
        )

    for key, value in inputs.items():
        has_wildcards = any("*" in uri for uri in value)
        keep_directories = KeepDirectories.SUFFIX.value if has_wildcards else False
        config_step.inputs[key] = Input(
            name=key, default=value, keep_directories=keep_directories
        )
    return config_step


def generate_config(
    *,
    relative_source_path: str,
    step: str,
    image: str,
    parameters: ParameterDict,
    inputs: InputDict,
) -> Config:
    step_obj = generate_step(
        relative_source_path=relative_source_path,
        step=step,
        image=image,
        parameters=parameters,
        inputs=inputs,
    )
    config = Config()
    config.steps[step_obj.name] = step_obj
    return config


def get_source_relative_path(source_path: str, config_path: str) -> str:
    """Return path of a source file relative to config file path

    :param source_path: Path of the Python source code file
    :param config_path: Path of the valohai.yaml config file

    :return: Path of the source code file relative to the config file
    Example:
        config_file_path: /herpderp/valohai.yaml
        source_file_path: /herpderp/somewhere/underneath/test.py
        return: ./somewhere/underneath/test.py
    Ultimately used for creating command with correct relative path in valohai.yaml:
        python ./somewhere/underneath/test.py {parameters}
    """
    relative_source_dir = os.path.relpath(
        os.path.dirname(os.path.abspath(source_path)),
        os.path.dirname(os.path.abspath(config_path)),
    )
    return os.path.join(relative_source_dir, os.path.basename(source_path))


def parse_config_from_source(source_path: str, config_path: str) -> Config:
    with open(source_path) as source_file:
        parsed = parse(source_file.read())
        if not parsed.step:
            raise ValueError("Source is missing a call to valohai.prepare()")
        relative_source_path = get_source_relative_path(source_path, config_path)
        return generate_config(
            relative_source_path=relative_source_path,
            step=parsed.step,
            image=DEFAULT_DOCKER_IMAGE if parsed.image is None else parsed.image,
            parameters=parsed.parameters,
            inputs=parsed.inputs,
        )


def get_parameter_type_name(name: str, value: Any) -> str:
    if isinstance(value, bool):
        return "flag"
    if isinstance(value, float):
        return "float"
    if isinstance(value, int):
        return "integer"
    if isinstance(value, str):
        return "string"

    raise ValueError(
        "Unrecognized parameter type for %s=%s. Supported Python types are float, int, string and bool."
        % (name, value)
    )



from collections import OrderedDict

import yaml
from valohai_yaml.objs.config import Config

# https://stackoverflow.com/questions/42518067/how-to-use-ordereddict-as-an-input-in-yaml-dump-or-yaml-safe-dump
yaml.add_representer(
    OrderedDict,
    lambda dumper, data: dumper.represent_mapping(
        "tag:yaml.org,2002:map", data.items()
    ),
)


def config_to_yaml(config: Config) -> str:
    """Serialize Valohai Config to YAML

    :param config: valohai_yaml.objs.Config object
    """

    return yaml.dump(config.serialize(), default_flow_style=False)



import os
from typing import Optional

from valohai.config import is_running_in_valohai
from valohai.consts import (
    VH_LOCAL_CONFIG_DIR,
    VH_LOCAL_INPUTS_DIR,
    VH_LOCAL_OUTPUTS_DIR,
    VH_LOCAL_REPOSITORY_DIR,
)
from valohai.internals import global_state
from valohai.internals.guid import get_execution_guid


def get_config_path() -> str:
    return os.environ.get(
        "VH_CONFIG_DIR",
        "/valohai/config" if is_running_in_valohai() else VH_LOCAL_CONFIG_DIR,
    )


def get_inputs_path(input_name: Optional[str] = None) -> str:
    if is_running_in_valohai():
        path = os.environ.get("VH_INPUTS_DIR", "/valohai/inputs")
    else:
        path = os.environ.get(
            "VH_INPUTS_DIR", os.path.join(VH_LOCAL_INPUTS_DIR, global_state.step_name)
        )

    if input_name:
        return os.path.join(path, input_name)
    return path


def get_outputs_path() -> str:
    if is_running_in_valohai():
        return os.environ.get("VH_OUTPUTS_DIR", "/valohai/outputs")
    else:
        return os.environ.get(
            "VH_OUTPUTS_DIR",
            os.path.join(
                VH_LOCAL_OUTPUTS_DIR, get_execution_guid(), global_state.step_name
            ),
        )


def get_repository_path() -> str:
    return os.environ.get(
        "VH_REPOSITORY_DIR",
        "/valohai/repository" if is_running_in_valohai() else VH_LOCAL_REPOSITORY_DIR,
    )


def get_inputs_config_path() -> str:
    return os.path.join(get_config_path(), "inputs.json")


def get_parameters_config_path() -> str:
    return os.path.join(get_config_path(), "parameters.json")



__version__ = "0.1.8"

import papi

from .inputs import inputs
from .metadata import logger
from .outputs import outputs
from .parameters import parameters
from .utils import prepare

Pipeline = papi.Papi

__all__ = ["inputs", "logger", "outputs", "parameters", "prepare", "Pipeline"]



import glob
import os
import pathlib
from stat import S_IREAD, S_IRGRP, S_IROTH
from typing import Set, Union


def set_file_read_only(path: str) -> None:
    os.chmod(path, S_IREAD | S_IRGRP | S_IROTH)


def get_glob_pattern(source: str) -> str:
    # Path is transformed into glob supported pattern. "example" -> "example/*"
    if os.path.isdir(source):
        return f"{source.rstrip('/')}/*"
    return source


def expand_globs(sources: Union[str, list], preprocessor=lambda s: s) -> Set[str]:
    if isinstance(sources, str):
        sources = [sources]
    files_to_compress = set()
    for source_path in sources:
        source_path = preprocessor(source_path)
        glob_pattern = get_glob_pattern(source_path)
        for file_path in glob.glob(glob_pattern):
            if os.path.isfile(file_path):
                files_to_compress.add(file_path)
    if not files_to_compress:
        raise ValueError(f"No files to compress at {sources}")
    return files_to_compress


def get_canonical_extension(pathname: str) -> str:
    """
    Get a canonicalized extension from a pathname. Correctly handles e.g. `tar.gz`.
    """
    return "".join(pathlib.Path(pathname).suffixes).lower()



import ast
from collections import namedtuple
from typing import Any, Dict, List, Optional


def is_module_function_call(node: ast.Call, module: str, function: str) -> bool:
    try:
        return node.func.attr == function and node.func.value.id == module  # type: ignore
    except AttributeError:
        return False


class PrepareParser(ast.NodeVisitor):
    """Parses .py file for Valohai inputs, parameters, step name

    Using AST parser, visits all method calls of a Python file.

    If a call to valohai.prepare() method is found, iterate through
    it's arguments and look for inputs, parameters and a step name.

    All possible ways to call prepare() are not supported

    Works:
        valohai.prepare(parameters={"param1": "foobar"})

    Works:
        parameters={"param1": "foobar"}
        valohai.prepare(parameters=parameters)

    Fails:
        import valohai as herpderp
        herpderp.prepare(parameters={"param1": "foobar"})

    Fails:
        from valohai import prepare
        prepare(parameters={"param1": "foobar"})

    Fails:
        valohai.prepare(parameters=get_parameters())
    """

    assignments: Dict[str, Any]  # TODO: embetter type
    parameters: Dict[str, Any]  # TODO: embetter type
    inputs: Dict[str, Any]  # TODO: embetter type
    step: Optional[str]

    def __init__(self) -> None:
        self.assignments = {}
        self.parameters = {}
        self.inputs = {}
        self.step = None
        self.image = None

    def visit_Assign(self, node: ast.Assign) -> None:
        try:
            self.assignments[node.targets[0].id] = ast.literal_eval(node.value)  # type: ignore
        except (ValueError, AttributeError):
            # We don't care about assignments that can't be literal_eval():ed
            pass

    def visit_Call(self, node: ast.Call) -> None:
        if is_module_function_call(node, "valohai", "prepare"):
            self.process_valohai_prepare_call(node)

    def process_valohai_prepare_call(self, node: ast.Call) -> None:
        self.step = "default"
        if hasattr(node, "keywords"):
            for key in node.keywords:
                if key.arg == "default_parameters":
                    self.process_default_parameters_arg(key)
                elif key.arg == "default_inputs":
                    self.process_default_inputs_arg(key)
                elif key.arg == "step":
                    self.step = ast.literal_eval(key.value)
                elif key.arg == "image":
                    self.image = ast.literal_eval(key.value)

    def process_default_inputs_arg(self, key: ast.keyword) -> None:
        if isinstance(key.value, ast.Name) and key.value.id in self.assignments:
            self.inputs = {
                key: value if isinstance(value, List) else [value]
                for key, value in self.assignments[key.value.id].items()
            }
        elif isinstance(key.value, ast.Dict):
            self.inputs = {
                key: value if isinstance(value, List) else [value]
                for key, value in ast.literal_eval(key.value).items()
            }
        else:
            raise NotImplementedError()

    def process_default_parameters_arg(self, key: ast.keyword) -> None:
        if isinstance(key.value, ast.Name) and key.value.id in self.assignments:
            self.parameters = self.assignments[key.value.id]
        elif isinstance(key.value, ast.Dict):
            self.parameters = ast.literal_eval(key.value)
        else:
            raise NotImplementedError()


def parse(source):
    tree = ast.parse(source)
    parser = PrepareParser()
    parser.visit(tree)
    result = namedtuple("result", ["step", "parameters", "inputs", "image"])
    return result(
        step=parser.step,
        parameters=parser.parameters,
        inputs=parser.inputs,
        image=parser.image,
    )



import ast
import os
import re

import setuptools


def get_version():
    with open(
        os.path.join(os.path.dirname(__file__), "valohai", "__init__.py")
    ) as infp:
        match = re.search("__version__ = (.+?)$", infp.read(), re.M)
        if not match:
            raise ValueError("No version could be found")
        return ast.literal_eval(match.group(1))


setuptools.setup(
    name="valohai-utils",
    version=get_version(),
    author="Valohai",
    author_email="hait@valohai.com",
    license="MIT",
    packages=setuptools.find_packages(include=("valohai*",)),
    install_requires=[
        "tqdm",
        "requests",
        "valohai-yaml>=0.13.0",
        "valohai-papi>=0.1.1",
    ],
    python_requires=">=3.6",
)



import os
import shutil
import tempfile
from contextlib import ExitStack
from tarfile import TarFile, TarInfo
from typing import IO, List, Optional, Union
from zipfile import ZipFile, ZipInfo


class File:
    parent_file: Optional["File"] = None

    def open(self) -> IO[bytes]:
        raise NotImplementedError("...")

    def read(self) -> bytes:
        with self.open() as f:
            return f.read()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    @property
    def name(self) -> str:
        raise NotImplementedError("...")

    @property
    def extension(self) -> str:
        return os.path.splitext(self.name)[1]


class FileOnDisk(File):
    _name: str
    path: str
    dir_entry: Optional[os.DirEntry]

    def __init__(
        self, name: str, path: str, dir_entry: Optional[os.DirEntry] = None
    ) -> None:
        self._name = name
        self.path = path
        self.dir_entry = dir_entry

    def open(self) -> IO[bytes]:
        return open(self.path, "rb")  # noqa: SIM115

    @property
    def name(self) -> str:
        return self._name


class FileInContainer(File):
    _concrete_path: Optional[str] = None

    def open_concrete(self, delete: bool = True) -> IO[bytes]:
        if self._concrete_path and os.path.isfile(self._concrete_path):
            return open(self._concrete_path, "rb")  # noqa: SIM115
        tf = tempfile.NamedTemporaryFile(suffix=self.extension, delete=delete)
        self.extract(tf)
        tf.seek(0)
        self._concrete_path = tf.name
        return tf

    def extract(self, destination: Union[str, IO]) -> None:
        if isinstance(destination, str):
            destination = open(destination, "wb")  # noqa: SIM115
            should_close = True
        else:
            should_close = False
        try:
            self._do_extract(destination)
        finally:
            if should_close:
                destination.close()

    def _do_extract(self, destination: IO) -> None:
        # if a file has a better idea how to write itself into an IO, this is the place
        with self.open() as f:
            shutil.copyfileobj(f, destination)


class FileInZip(FileInContainer):
    parent_file: FileOnDisk
    zipfile: ZipFile
    zipinfo: ZipInfo

    def __init__(
        self, parent_file: FileOnDisk, zipfile: ZipFile, zipinfo: ZipInfo
    ) -> None:
        self.parent_file = parent_file
        self.zipfile = zipfile
        self.zipinfo = zipinfo

    def open(self) -> IO[bytes]:
        return self.zipfile.open(self.zipinfo, "r")

    @property
    def name(self) -> str:
        return os.path.join(
            os.path.dirname(self.parent_file.name), self.zipinfo.filename
        )

    @property
    def path(self) -> str:
        return os.path.join(
            os.path.dirname(self.parent_file.path), self.zipinfo.filename
        )


class FileInTar(FileInContainer):
    parent_file: FileOnDisk
    tarfile: TarFile
    tarinfo: TarInfo

    def __init__(
        self, parent_file: FileOnDisk, tarfile: TarFile, tarinfo: TarInfo
    ) -> None:
        self.parent_file = parent_file
        self.tarfile = tarfile
        self.tarinfo = tarinfo

    def open(self) -> IO[bytes]:
        fp = self.tarfile.extractfile(self.tarinfo)
        if not fp:
            raise ValueError(f"extractfile() returned None for {self.tarinfo}")
        return fp

    @property
    def name(self) -> str:
        return os.path.join(os.path.dirname(self.parent_file.name), self.tarinfo.path)


def find_files_in_zip(vr: "VFS", df: FileOnDisk) -> None:
    zf = ZipFile(df.path)
    df.zipfile = zf  # type: ignore
    vr.exit_stack.callback(zf.close)
    vr.files.extend(
        [
            FileInZip(parent_file=df, zipfile=zf, zipinfo=zinfo)
            for zinfo in zf.infolist()
        ]
    )


def find_files_in_tar(vr: "VFS", df: FileOnDisk) -> None:
    tf = TarFile.open(df.path)
    df.tarfile = tf  # type: ignore
    vr.exit_stack.callback(tf.close)
    vr.files.extend(
        [
            FileInTar(parent_file=df, tarfile=tf, tarinfo=tinfo)
            for tinfo in tf.getmembers()
            if tinfo.isreg()
        ]
    )


class VFS:
    files: List[File]

    def __init__(self) -> None:
        self.files = []
        self.exit_stack = ExitStack()

    def __enter__(self) -> "VFS":
        return self

    def __exit__(self, *exc_details) -> None:
        self.exit_stack.__exit__(*exc_details)


def add_disk_file(
    vfs: VFS,
    name: str,
    path: str,
    dir_entry: Optional[os.DirEntry] = None,
    process_archives: bool = False,
) -> None:
    disk_file = FileOnDisk(name=name, path=path, dir_entry=dir_entry)
    if process_archives:
        extension = disk_file.extension.lower()
        if extension == ".zip":
            find_files_in_zip(vfs, disk_file)
            return
        elif extension in (".tar", ".tar.gz", ".tar.bz2", ".tar.xz"):
            find_files_in_tar(vfs, disk_file)
            return
    vfs.files.append(disk_file)


def find_files(vfs: VFS, root: str, *, process_archives: bool) -> None:
    dent: os.DirEntry

    def _walk(path):
        for dent in os.scandir(path):
            if dent.is_dir():
                _walk(dent.path)
                continue
            if not dent.is_file():
                continue
            add_disk_file(
                vfs,
                name=os.path.relpath(dent.path, root),
                path=dent.path,
                dir_entry=dent,
                process_archives=process_archives,
            )

    _walk(root)



from valohai.internals import global_state
from valohai.internals.parameters import load_parameter, supported_types


class Parameter:
    def __init__(self, name: str, default: supported_types = None) -> None:
        self.name = str(name)
        self.default = default

    @property
    def value(self) -> supported_types:
        if self.name in global_state.parsed_parameters:
            return global_state.parsed_parameters[self.name]
        return load_parameter(self.name, self.default)

    @value.setter
    def value(self, value: supported_types):
        global_state.parsed_parameters[self.name] = value


parameters = Parameter
